/*! \mainpage

\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim

 * This file contains code made by Sylvain Paris under
 * supervision of Franois Sillion for his PhD work with ARTIS
 * project. ARTIS is a research project in the GRAVIR/IMAG laboratory,
 * a joint unit of CNRS, INPG, INRIA and UJF.
 *
 *  This file provides the graph implementation as first used in the
 * ACCV'04 paper <i>A Surface Reconstruction Method Using Global Graph
 * Cut Optimization</i> from Paris, Sillion and Quan. If you use this
 * code, please let us know. We are interested in anything related to
 * this work.
 *
 *  The code is in a "debug state": it is very verbose in case of
 * error. Therefore it is clearly not as efficient as it could be.
 *
 *  To run this code, three other header files are required: array.h,
 * geom.h and msg_stream.h
 *
 * Source code can be downloaded from <a
 * href="http://artis.imag.fr/Members/Sylvain.Paris/index.html#code">Sylvain
 * Paris'page</a>.
 *
 *  Here is an example of code:
 * \code

     #include "pushrelabel_non_linear_grid_3D.h"

     // ...
 
     typedef float                                 real_type;
     typedef unsigned int                          size_type;
     typedef PushRelabel_non_linear_grid_3D<float> graph_type;

     const size_type x_size = 100;
     const size_type y_size = 200;
     const size_type z_size = 300;
     
     Array_3D<real_type> consistency(x_size,y_size,z_size);
     Array_3D<bool>      domain(x_size,y_size,z_size);
     
     // ...
     // Some code determining the domain.
     // ...
     // Some code computing the consistency.
     // ...
     
     Array_2D<real_type> x_smoothness(x_size,y_size);
     Array_2D<real_type> y_smoothness(x_size,y_size);

     // ...
     // Some code computing the smoothness values.
     // ...

     graph_type graph(consistency,domain,x_smoothness,y_smoothness);

     graph.max_flow(); // This call takes most the computing time.

     Array_2D<size_type> cut;
     graph.min_cut(&cut);

     // Now, the 2D array contains the values of the cut i.e. z = cut(x,y).
  * \endcode
  *
  */

#ifndef __PUSHRELABEL_NON_LINEAR_GRID_3D__
#define __PUSHRELABEL_NON_LINEAR_GRID_3D__



#include <algorithm>
#include <deque>
#include <iterator>

#ifndef WITHOUT_LIMITS
#include <limits>
#endif

#include <list>
#include <vector>

#include "array.h"
#include "geom.h"
#include "msg_stream.h"



namespace Graph_cut{

  

/*

###################################
#  PushRelabel_non_linear_grid_3D #
###################################

*/



//! Class containing functions to compute a flow
//! on a graph aligned on complex 2D grid.
/*!
  Optimized so as to use as few memory as possible : combinatorial
  information (adjacency) are compute on the fly at each request.
*/
  template<typename Real>
  class PushRelabel_non_linear_grid_3D{

  public:
  
    //@{
    //! Useful typedef.
 
    typedef Real                      real_type;
    typedef Array_3D<bool>::size_type size_type;
    //@}

    //! Constructor.
    PushRelabel_non_linear_grid_3D(
      const Array_3D<real_type>& cost_edge_cap,
      const Array_3D<bool>&      dom,
      const Array_2D<real_type>& x_derivative_edge_cap,
      const Array_2D<real_type>& y_derivative_edge_cap,
      const real_type            eps = 1e-3);

    //! Compute the maximum flow on the graph.
    real_type max_flow();

    //! Compute the minimum cut once the maximum flow is known.
    void min_cut(Array_2D<size_type>* const cut) const;

//  private:

    bool DBG_LOGGING;
  
    //@{
    //! Useful typedef.
  
    typedef Geometry::Vec3<unsigned int> position_type;
    typedef unsigned int                 index_type;
    typedef unsigned int                 label_type;
    //@}

    //@{
    //! Direction used to name an edge by the pair (node,direction).
    typedef char direction_type;
  
    static const direction_type X_MIN_DIR       = 1;
    static const direction_type X_MAX_DIR       = 2;

    static const direction_type Y_MIN_DIR       = 3;
    static const direction_type Y_MAX_DIR       = 4;

    static const direction_type Z_MIN_x_min_DIR = 5;
    static const direction_type Z_MIN_x_max_DIR = 6;
    static const direction_type Z_MIN_y_min_DIR = 7;
    static const direction_type Z_MIN_y_max_DIR = 8;
  
    static const direction_type Z_MAX_x_min_DIR = 9;
    static const direction_type Z_MAX_x_max_DIR = 10;
    static const direction_type Z_MAX_y_min_DIR = 11;
    static const direction_type Z_MAX_y_max_DIR = 12;
    //@}

    //@{
    //! Position to name a node of coordinate (x;y).
    typedef char location_type;

    static const location_type CENTER_NODE      = 100;
    static const location_type Z_MAX_x_min_NODE = 101;
    static const location_type Z_MAX_x_max_NODE = 102;
    static const location_type Z_MAX_y_min_NODE = 103;
    static const location_type Z_MAX_y_max_NODE = 104;
    //@}

    //! Node type
    struct node_type{
      position_type pos;
      location_type loc;

      inline node_type(){}
      inline node_type(const position_type& p,const location_type l):pos(p),loc(l){}
      inline bool operator==(const node_type& n) const {
	return (pos==n.pos)&&(loc==n.loc);
      }
      inline bool operator!=(const node_type& n) const {
	return !((pos==n.pos)&&(loc==n.loc));
      }
    };

    //! Threshold to test zero equality.
    const real_type epsilon;

    //! Number of nodes in the graph.
    size_type nb_nodes;

    //! Number of consecutive \e relabel steps.
    size_type nb_relabel;
  
    //! Max capacity of the \e cost edges.
    Array_3D<real_type> cost_edge_capacity;

    //! Definition of the optimization domain. Graph is only
    //! defined on this domain.
    const Array_3D<bool>& domain;

    //@{
    //! Max capacity of the derivative penalty edges.
  
    Array_2D<real_type> x_major_derivative_edge_capacity;
    Array_2D<real_type> x_minor_derivative_edge_capacity;

    Array_2D<real_type> y_major_derivative_edge_capacity;
    Array_2D<real_type> y_minor_derivative_edge_capacity;
    //@}

    //@{
    //! Node labels.

    Array_3D<index_type> label_CENTER;
    Array_3D<index_type> label_Z_MAX_x_min;
    Array_3D<index_type> label_Z_MAX_x_max;
    Array_3D<index_type> label_Z_MAX_y_min;
    Array_3D<index_type> label_Z_MAX_y_max;
    //@}

    //@{
    //! Flow in the \e cost edges.
    /*! Sign refers to the flow direction. */
  
    Array_3D<real_type> cost_edge_Z_MAX_x_min_flow;
    Array_3D<real_type> cost_edge_Z_MAX_x_max_flow;
    Array_3D<real_type> cost_edge_Z_MAX_y_min_flow;
    Array_3D<real_type> cost_edge_Z_MAX_y_max_flow;
  
    Array_3D<real_type> cost_edge_Z_MAX_MAX_x_min_flow;
    Array_3D<real_type> cost_edge_Z_MAX_MAX_x_max_flow;
    Array_3D<real_type> cost_edge_Z_MAX_MAX_y_min_flow;
    Array_3D<real_type> cost_edge_Z_MAX_MAX_y_max_flow;
    //@}

    //@{
    //! Flow in the \e penalty edges.
    /*! Sign refers to the flow direction. */
  
    Array_3D<real_type> x_major_derivative_edge_flow;
    Array_3D<real_type> y_major_derivative_edge_flow;
  
    Array_3D<real_type> x_minor_derivative_edge_flow;
    Array_3D<real_type> y_minor_derivative_edge_flow;
    //@}

    //! Bottom boundary the optimization domain.
    Array_2D<size_type> min_bound;

    //! Top boundary of the optimization domain.
    Array_2D<size_type> max_bound;

    //! List of the active nodes.
    std::deque<index_type> active_node_index;
  
    //! Source node.
    const node_type source;

    //! Think node.
    const node_type target;

    //@{
    //! Max size of the optimization domain.

    const size_type x_size,y_size,z_size;
    //@}


    //! Compute an index for all the 3 coordinates to spare space.
    inline void to_index(const node_type& p,
			 size_type* const i) const;

    //! Reverse transformation of to_index().
    inline void from_index(node_type* const p,
			   const size_type i) const;

    //@{
    //! Manage the labels.
  
    inline const label_type& label(const node_type& node) const;
    inline label_type& label(const node_type& node);
    //@}
  
    //! Gives all the neighbors.
    inline void neighbors(const node_type& node,
			  node_type* const result,
			  size_type* const n_res) const;
  
    //! Gives the adjacent nodes that are reachable through a non saturated edge
    inline void adjacent_nodes(const node_type& node,
			       node_type* const result,
			       size_type* const n_res) const;

    //! Gives the max capacity of an edge named by (node,direction).
    inline real_type edge_capacity(const node_type&     from,
				   const direction_type dir) const;

    //! Gives the flow in an edge named by (node,direction).
    inline real_type edge_flow(const node_type&     from,
			       const direction_type dir) const;


    //! Gives the residual capacity of an edge named by (node,direction).
    inline real_type edge_residual(const node_type&     from,
				   const direction_type dir) const;

    //! Gives excess water in a node.
    inline real_type node_excess(const node_type& node) const;
  

    //! Adds flow to an edge.
    inline void add_flow(const node_type&     from,
			 const direction_type dir,
			 const real_type      flow);

    //! Gives the direction of an edge named by (source,target).
    inline direction_type direction(const node_type& from,
				    const node_type& to) const;

    //! \e relabel step of the algorithm.
    inline label_type relabel(const node_type& node);

    //! Global update of the labels.
    void global_relabel();

    //! \e discharge step of the algorithm.
    inline bool discharge(const node_type& node);

    //! Adds an active node.
    inline void add_active_node(const node_type& node);

    //! Gets the next node to handle.
    inline void get_next_node(node_type* const node,
			      bool* const      no_more_node);

    //! Algorithm defined test.
    inline bool is_active(const node_type& node) const;

    //! Algorithm defined test.
    inline bool is_admissible(const node_type& from,
			      const node_type& to) const;


  };




  
/*
  
#############################################
#############################################
#############################################
######                                 ######
######   I M P L E M E N T A T I O N   ######
######                                 ######
#############################################
#############################################
#############################################
  
*/



  
/*

########################################
# class PushRelabel_non_linear_grid_3D #
########################################

*/





  /*! You have to provide the consistency value for each voxel in \p
   *  cost_edge_cap which is a 3D array (remember the final cut is
   *  parameterized as z = h(x,y)). \p dom is a 3D array with boolean
   *  values indacting if the (x,y,z)-voxel is part of optimization
   *  domain. \p x_derivative_edge_cap is 2D array containing the
   *  smoothing value between (x,y)-column and (x+1,y)-column. \p
   *  y_derivative_edge_cap is the equivalent part for (x,y) and
   *  (x,y+1). \p eps is a value such that any value smallest than eps
   *  is considered to be zero.
   */
  template<typename Real>
  PushRelabel_non_linear_grid_3D<Real>::
  PushRelabel_non_linear_grid_3D(
    const Array_3D<real_type>& cost_edge_cap,
    const Array_3D<bool>&      dom,
    const Array_2D<real_type>& x_derivative_edge_cap,
    const Array_2D<real_type>& y_derivative_edge_cap,
    const real_type            eps)
    :epsilon(eps),
     nb_nodes(2),
     nb_relabel(0),
     //cost_edge_capacity(cost_edge_cap),
     domain(dom),
     DBG_LOGGING(false),
   
     x_major_derivative_edge_capacity(x_derivative_edge_cap),
     x_minor_derivative_edge_capacity(x_derivative_edge_cap),

     y_major_derivative_edge_capacity(y_derivative_edge_cap),
     y_minor_derivative_edge_capacity(y_derivative_edge_cap),
   
     cost_edge_Z_MAX_x_min_flow(dom.x_size(),dom.y_size(),dom.z_size(),0),
     cost_edge_Z_MAX_x_max_flow(dom.x_size(),dom.y_size(),dom.z_size(),0),
     cost_edge_Z_MAX_y_min_flow(dom.x_size(),dom.y_size(),dom.z_size(),0),
     cost_edge_Z_MAX_y_max_flow(dom.x_size(),dom.y_size(),dom.z_size(),0),
   
     cost_edge_Z_MAX_MAX_x_min_flow(dom.x_size(),dom.y_size(),dom.z_size(),0),
     cost_edge_Z_MAX_MAX_x_max_flow(dom.x_size(),dom.y_size(),dom.z_size(),0),
     cost_edge_Z_MAX_MAX_y_min_flow(dom.x_size(),dom.y_size(),dom.z_size(),0),
     cost_edge_Z_MAX_MAX_y_max_flow(dom.x_size(),dom.y_size(),dom.z_size(),0),
   
     x_major_derivative_edge_flow(dom.x_size(),dom.y_size(),dom.z_size()+1,0),
     x_minor_derivative_edge_flow(dom.x_size(),dom.y_size(),dom.z_size()+1,0),
     y_major_derivative_edge_flow(dom.x_size(),dom.y_size(),dom.z_size()+1,0),
     y_minor_derivative_edge_flow(dom.x_size(),dom.y_size(),dom.z_size()+1,0),

     label_CENTER(dom.x_size(),dom.y_size(),dom.z_size()+1,0),
     label_Z_MAX_x_min(dom.x_size(),dom.y_size(),dom.z_size()+1,0),
     label_Z_MAX_x_max(dom.x_size(),dom.y_size(),dom.z_size()+1,0),
     label_Z_MAX_y_min(dom.x_size(),dom.y_size(),dom.z_size()+1,0),
     label_Z_MAX_y_max(dom.x_size(),dom.y_size(),dom.z_size()+1,0),
   
     min_bound(dom.x_size(),dom.y_size()),
     max_bound(dom.x_size(),dom.y_size()),
   
     x_size(dom.x_size()),
     y_size(dom.y_size()),
     z_size(dom.z_size()),

#ifndef WITHOUT_LIMITS
     source(position_type(std::numeric_limits<size_type>::max(),
			  std::numeric_limits<size_type>::max(),
			  std::numeric_limits<size_type>::max()),CENTER_NODE),
     target(position_type(std::numeric_limits<size_type>::max()-1,
			  std::numeric_limits<size_type>::max()-1,
			  std::numeric_limits<size_type>::max()-1),CENTER_NODE)
#else
     source(position_type(static_cast<size_type>(4294967295),
			  static_cast<size_type>(4294967295),
			  static_cast<size_type>(4294967295)),CENTER_NODE),
     target(position_type(static_cast<size_type>(4294967294),
			  static_cast<size_type>(4294967294),
			  static_cast<size_type>(4294967294)),CENTER_NODE)
#endif
  {

    // Computes the top and bottom boundaries of the optimization domain.
    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
      
	size_type top = 0;
	bool found = false;
    
	for(top=0;(top<z_size)&&(!found);top++){
	  found = domain(x,y,top);
	}
	min_bound(x,y) = top-1; // -1 because we are one step too far.
      
	add_active_node(node_type(position_type(x,y,min_bound(x,y)),CENTER_NODE));
      
	size_type bottom = 0;
	found = false;
	for(bottom=0;(bottom<z_size)&&(!found);bottom++){
	  found = domain(x,y,z_size-1-bottom);
	}
	max_bound(x,y) = z_size - bottom + 1;
      // +1 because we are one step too far
      //and we takethe  MAX node of the edge [MIN;MAX]

      }//end of for y
    }// end of for x

//    for(size_type x=0;x<x_size;x++){
//      for(size_type y=0;y<y_size;y++){
//        if ((max_bound(x,y)-min_bound(x,y))<=3){
//  	std::cout<<x<<"\t"<<y<<"\t"
//  		 <<VALUE_OF(min_bound(x,y))<<"\t"
//  		 <<VALUE_OF(max_bound(x,y))<<std::endl;
//        }
//      }
//    }

//    exit(0);


    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
	
	x_major_derivative_edge_capacity(x,y) *= 1.0;
	x_minor_derivative_edge_capacity(x,y) *= 1.0;
      
	y_major_derivative_edge_capacity(x,y) *= 1.0;
	y_minor_derivative_edge_capacity(x,y) *= 1.0;
        
	/*      
	x_major_derivative_edge_capacity(x,y) *= 2.0;
        x_minor_derivative_edge_capacity(x,y) *= .0;

        y_major_derivative_edge_capacity(x,y) *= 2.0;
        y_minor_derivative_edge_capacity(x,y) *= .0;
        */
      }
    }

    real_type max_minor = 0;
    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
	if (max_minor < x_minor_derivative_edge_capacity(x,y)){
	  max_minor = x_minor_derivative_edge_capacity(x,y);
	}
	if (max_minor < y_minor_derivative_edge_capacity(x,y)){
	  max_minor = y_minor_derivative_edge_capacity(x,y);
	}
      }
    }
  
    cost_edge_capacity.resize(x_size,y_size,z_size);
  
    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
	for(size_type z=0;z<z_size;z++){
	  cost_edge_capacity(x,y,z) = cost_edge_cap(x,y,z)/4 + max_minor;

	  if (domain(x,y,z)){
	    nb_nodes += 12;
	  }
	}
      }
    }

//    std::cout<<VALUE_OF(max_minor)<<std::endl;
  }


  template<typename Real>
  void 
  PushRelabel_non_linear_grid_3D<Real>::
  to_index(const node_type& p,
	   index_type* const i) const{

    if (p==source) {
#ifndef WITHOUT_LIMITS
      *i = std::numeric_limits<index_type>::max();
#else
      *i = static_cast<size_type>(4294967295);
#endif
    }
    else if (p==target){
#ifndef WITHOUT_LIMITS
      *i = std::numeric_limits<index_type>::max()-1;
#else
      *i = static_cast<size_type>(4294967294);
#endif
    }
    else{
    
      *i = ((p.pos.x() * y_size * z_size) + (p.pos.y() * z_size) + p.pos.z()) * 5;
      switch(p.loc){
      case CENTER_NODE:
	break;

      case Z_MAX_x_min_NODE:
	*i += 1;
	break;

      case Z_MAX_x_max_NODE:
	*i += 2;
	break;

      case Z_MAX_y_min_NODE:
	*i += 3;
	break;

      case Z_MAX_y_max_NODE:
	*i += 4;
	break;  

      default:
	Message::error<<"to_index: unknown value of p.loc"<<Message::done;
      }
    
    }
  }


  template<typename Real>
  void
  PushRelabel_non_linear_grid_3D<Real>::
  from_index(node_type* const p,
	     const index_type i) const{

#ifndef WITHOUT_LIMITS
    if (i==std::numeric_limits<index_type>::max()){
#else
    if (i==static_cast<size_type>(4294967295)){
#endif
      *p = source;
    }
#ifndef WITHOUT_LIMITS
    else if (i==std::numeric_limits<index_type>::max()-1){
#else
    else if (i==static_cast<size_type>(4294967294)){
#endif
      *p = target;
    }
    else{
      switch(i%5){
      case 0:
	p->loc = CENTER_NODE;
	break;
      case 1:
	p->loc = Z_MAX_x_min_NODE;
	break;
      case 2:
	p->loc = Z_MAX_x_max_NODE;
	break;
      case 3:
	p->loc = Z_MAX_y_min_NODE;
	break;
      case 4:
	p->loc = Z_MAX_y_max_NODE;
	break;
      }

      const index_type j = i/5;
      p->pos.x() = j / (y_size * z_size);
      p->pos.y() = (j / z_size) -  (p->pos.x() * y_size);
      p->pos.z() = j % z_size;
    }
  }


  template<typename Real>
  const typename PushRelabel_non_linear_grid_3D<Real>::label_type&
  PushRelabel_non_linear_grid_3D<Real>::label(const node_type& node) const{
  
    switch(node.loc){
    case CENTER_NODE:      return label_CENTER[node.pos];
    case Z_MAX_x_min_NODE: return label_Z_MAX_x_min[node.pos];
    case Z_MAX_x_max_NODE: return label_Z_MAX_x_max[node.pos];
    case Z_MAX_y_min_NODE: return label_Z_MAX_y_min[node.pos];
    case Z_MAX_y_max_NODE: return label_Z_MAX_y_max[node.pos];
    }

    Message::error<<"label : non expected exit"<<Message::done;
    return label_CENTER[node.pos];
  }


  template<typename Real>
  typename PushRelabel_non_linear_grid_3D<Real>::label_type&
  PushRelabel_non_linear_grid_3D<Real>::label(const node_type& node){
  
    switch(node.loc){
    case CENTER_NODE:      return label_CENTER[node.pos];
    case Z_MAX_x_min_NODE: return label_Z_MAX_x_min[node.pos];
    case Z_MAX_x_max_NODE: return label_Z_MAX_x_max[node.pos];
    case Z_MAX_y_min_NODE: return label_Z_MAX_y_min[node.pos];
    case Z_MAX_y_max_NODE: return label_Z_MAX_y_max[node.pos];
    }

    Message::error<<"label : non expected exit"<<Message::done;
    return label_CENTER[node.pos];
  }



/*!
  Pas de vrification pour les bords.
*/
  template<typename Real>
  typename PushRelabel_non_linear_grid_3D<Real>::real_type
  PushRelabel_non_linear_grid_3D<Real>::
  edge_capacity(const node_type&     from,
		const direction_type dir) const{

    switch(from.loc){
      // ########################### CENTER_NODE ###########################
    case CENTER_NODE:
    
      switch(dir){
      case X_MIN_DIR:
	return x_major_derivative_edge_capacity(from.pos.x()-1,from.pos.y());

      case X_MAX_DIR:
	return x_major_derivative_edge_capacity(from.pos.x(),from.pos.y());

      case Y_MIN_DIR:
	return y_major_derivative_edge_capacity(from.pos.x(),from.pos.y()-1);

      case Y_MAX_DIR:
	return y_major_derivative_edge_capacity(from.pos.x(),from.pos.y());  

      case Z_MIN_x_min_DIR: 
      case Z_MIN_x_max_DIR:
      case Z_MIN_y_min_DIR: 
      case Z_MIN_y_max_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z()-1);

      case Z_MAX_x_min_DIR:
      case Z_MAX_x_max_DIR:
      case Z_MAX_y_min_DIR:
      case Z_MAX_y_max_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());
      }

      Message::error<<"edge_capacity : bad end 1"<<Message::done;
    
      // ########################### Z_MAX_x_min_NODE ###########################  
    case Z_MAX_x_min_NODE:

      switch(dir){
      case X_MIN_DIR:
	return x_minor_derivative_edge_capacity(from.pos.x()-1,from.pos.y());

      case X_MAX_DIR:
	Message::error<<"edge_capacity : bad end 2.1"<<Message::done;

      case Y_MIN_DIR:
	Message::error<<"edge_capacity : bad end 2.2"<<Message::done;

      case Y_MAX_DIR:
	Message::error<<"edge_capacity : bad end 2.3"<<Message::done;

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_capacity : bad end 2.4"<<Message::done;
   
      case Z_MIN_x_max_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_capacity : bad end 2.5"<<Message::done;
   
      case Z_MIN_y_max_DIR:
	Message::error<<"edge_capacity : bad end 2.6"<<Message::done;

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_capacity : bad end 2.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_capacity : bad end 2.8"<<Message::done;

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_capacity : bad end 2.9"<<Message::done;
      }
    
      Message::error<<"edge_capacity : bad end 2"<<Message::done;
    
      // ########################### Z_MAX_x_max_NODE ###########################
    case Z_MAX_x_max_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_capacity : bad end 3.1"<<Message::done;

      case X_MAX_DIR:
	return x_minor_derivative_edge_capacity(from.pos.x(),from.pos.y());

      case Y_MIN_DIR:
	Message::error<<"edge_capacity : bad end 3.2"<<Message::done;

      case Y_MAX_DIR:
	Message::error<<"edge_capacity : bad end 3.3"<<Message::done;

      case Z_MIN_x_min_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());

      case Z_MIN_x_max_DIR: 
	Message::error<<"edge_capacity : bad end 3.4"<<Message::done;

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_capacity : bad end 3.5"<<Message::done;

      case Z_MIN_y_max_DIR: 
	Message::error<<"edge_capacity : bad end 3.6"<<Message::done;

      case Z_MAX_x_min_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_capacity : bad end 3.7"<<Message::done;

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_capacity : bad end 3.8"<<Message::done;;

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_capacity : bad end 3.9"<<Message::done;  
      }
    
      Message::error<<"edge_capacity : bad end 3"<<Message::done;
      return 0;

      // ########################### Z_MAX_y_min_NODE ###########################  
    case Z_MAX_y_min_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_capacity : bad end 4.1"<<Message::done;

      case X_MAX_DIR:
	Message::error<<"edge_capacity : bad end 4.2"<<Message::done;

      case Y_MIN_DIR:
	return y_minor_derivative_edge_capacity(from.pos.x(),from.pos.y()-1);

      case Y_MAX_DIR:
	Message::error<<"edge_capacity : bad end 4.3"<<Message::done;

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_capacity : bad end 4.4"<<Message::done;
   
      case Z_MIN_x_max_DIR:
	Message::error<<"edge_capacity : bad end 4.5"<<Message::done;      

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_capacity : bad end 4.6"<<Message::done;
   
      case Z_MIN_y_max_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_capacity : bad end 4.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_capacity : bad end 4.8"<<Message::done;

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_capacity : bad end 4.9"<<Message::done;

      case Z_MAX_y_max_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());
      }
    
      Message::error<<"edge_capacity : bad end 4"<<Message::done;
    
      // ########################### Z_MAX_y_max_NODE ###########################
    case Z_MAX_y_max_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_capacity : bad end 5.1"<<Message::done;

      case X_MAX_DIR:
	Message::error<<"edge_capacity : bad end 5.2"<<Message::done;

      case Y_MIN_DIR:
	Message::error<<"edge_capacity : bad end 5.3"<<Message::done;

      case Y_MAX_DIR:
	return y_minor_derivative_edge_capacity(from.pos.x(),from.pos.y());

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_capacity : bad end 5.4"<<Message::done;

      case Z_MIN_x_max_DIR: 
	Message::error<<"edge_capacity : bad end 5.5"<<Message::done;

      case Z_MIN_y_min_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());

      case Z_MIN_y_max_DIR: 
	Message::error<<"edge_capacity : bad end 5.6"<<Message::done;

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_capacity : bad end 5.7"<<Message::done;;

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_capacity : bad end 5.8"<<Message::done;

      case Z_MAX_y_min_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z());

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_capacity : bad end 5.9"<<Message::done;  
      }
    
      Message::error<<"edge_capacity : bad end 5"<<Message::done;
      return 0;
    }
  
  }

  template<typename Real>
  typename PushRelabel_non_linear_grid_3D<Real>::real_type
  PushRelabel_non_linear_grid_3D<Real>::
  edge_flow(const node_type&     from,
	    const direction_type dir) const{

    switch(from.loc){
      // ########################### CENTER_NODE ###########################
    case CENTER_NODE:
    
      switch(dir){
      case X_MIN_DIR:
	return -x_major_derivative_edge_flow(from.pos.x()-1,from.pos.y(),from.pos.z());
      
      case X_MAX_DIR:
	return x_major_derivative_edge_flow[from.pos];

      case Y_MIN_DIR:
	return -y_major_derivative_edge_flow(from.pos.x(),from.pos.y()-1,from.pos.z());
      
      case Y_MAX_DIR:
	return y_major_derivative_edge_flow[from.pos];

      case Z_MIN_x_min_DIR:
	return -cost_edge_Z_MAX_MAX_x_min_flow(from.pos.x(),from.pos.y(),from.pos.z()-1);
      
      case Z_MIN_x_max_DIR:
	return -cost_edge_Z_MAX_MAX_x_max_flow(from.pos.x(),from.pos.y(),from.pos.z()-1);

      case Z_MIN_y_min_DIR:
	return -cost_edge_Z_MAX_MAX_y_min_flow(from.pos.x(),from.pos.y(),from.pos.z()-1);
      
      case Z_MIN_y_max_DIR:
	return -cost_edge_Z_MAX_MAX_y_max_flow(from.pos.x(),from.pos.y(),from.pos.z()-1);

      case Z_MAX_x_min_DIR:
	return cost_edge_Z_MAX_x_min_flow[from.pos];
      
      case Z_MAX_x_max_DIR:
	return cost_edge_Z_MAX_x_max_flow[from.pos];

      case Z_MAX_y_min_DIR:
	return cost_edge_Z_MAX_y_min_flow[from.pos];
      
      case Z_MAX_y_max_DIR:
	return cost_edge_Z_MAX_y_max_flow[from.pos];
      }

      Message::error<<"edge_flow : bad end 1"<<Message::done;
      return 0;
    
      // ########################### Z_MAX_x_min_NODE ###########################  
    case Z_MAX_x_min_NODE:

      switch(dir){
      case X_MIN_DIR:
	return -x_minor_derivative_edge_flow(from.pos.x()-1,from.pos.y(),from.pos.z());

      case X_MAX_DIR:
	Message::error<<"edge_flow : bad end 2.1"<<Message::done;

      case Y_MIN_DIR:
	Message::error<<"edge_flow : bad end 2.2"<<Message::done;

      case Y_MAX_DIR:
	Message::error<<"edge_flow : bad end 2.3"<<Message::done;  

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_flow : bad end 2.4"<<Message::done;
   
      case Z_MIN_x_max_DIR:
	return -cost_edge_Z_MAX_x_min_flow[from.pos];

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_flow : bad end 2.5"<<Message::done;
   
      case Z_MIN_y_max_DIR:
	Message::error<<"edge_flow : bad end 2.6"<<Message::done;

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_flow : bad end 2.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	return cost_edge_Z_MAX_MAX_x_min_flow[from.pos];

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_flow : bad end 2.8"<<Message::done;

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_flow : bad end 2.9"<<Message::done;
      }
    
      Message::error<<"edge_flow : bad end 2"<<Message::done;
      return 0;
    
      // ########################### Z_MAX_x_max_NODE ###########################
    case Z_MAX_x_max_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_flow : bad end 3.1"<<Message::done;

      case X_MAX_DIR:
	return x_minor_derivative_edge_flow[from.pos];

      case Y_MIN_DIR:
	Message::error<<"edge_flow : bad end 3.2"<<Message::done;

      case Y_MAX_DIR:
	Message::error<<"edge_flow : bad end 3.3"<<Message::done;

      case Z_MIN_x_min_DIR:
	return -cost_edge_Z_MAX_x_max_flow[from.pos];

      case Z_MIN_x_max_DIR: 
	Message::error<<"edge_flow : bad end 3.4"<<Message::done;

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_flow : bad end 3.5"<<Message::done;

      case Z_MIN_y_max_DIR: 
	Message::error<<"edge_flow : bad end 3.6"<<Message::done;  

      case Z_MAX_x_min_DIR:
	return cost_edge_Z_MAX_MAX_x_max_flow[from.pos];

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_flow : bad end 3.7"<<Message::done;

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_flow : bad end 3.8"<<Message::done;

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_flow : bad end 3.9"<<Message::done;  
      }
    
      Message::error<<"edge_flow : bad end 3"<<Message::done;
      return 0;

      // ########################### Z_MAX_y_min_NODE ###########################  
    case Z_MAX_y_min_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_flow : bad end 4.1"<<Message::done;

      case X_MAX_DIR:
	Message::error<<"edge_flow : bad end 4.2"<<Message::done;

      case Y_MIN_DIR:
	return -y_minor_derivative_edge_flow(from.pos.x(),from.pos.y()-1,from.pos.z());


      case Y_MAX_DIR:
	Message::error<<"edge_flow : bad end 4.3"<<Message::done;  

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_flow : bad end 4.4"<<Message::done;
   
      case Z_MIN_x_max_DIR:
	Message::error<<"edge_flow : bad end 4.5"<<Message::done;

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_flow : bad end 4.6"<<Message::done;
   
      case Z_MIN_y_max_DIR:
	return -cost_edge_Z_MAX_y_min_flow[from.pos];

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_flow : bad end 4.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_flow : bad end 4.8"<<Message::done;

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_flow : bad end 4.9"<<Message::done;

      case Z_MAX_y_max_DIR:
	return cost_edge_Z_MAX_MAX_y_min_flow[from.pos];

      }
    
      Message::error<<"edge_flow : bad end 4"<<Message::done;
      return 0;
    
      // ########################### Z_MAX_y_max_NODE ###########################
    case Z_MAX_y_max_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_flow : bad end 5.1"<<Message::done;

      case X_MAX_DIR:
	Message::error<<"edge_flow : bad end 5.2"<<Message::done;

      case Y_MIN_DIR:
	Message::error<<"edge_flow : bad end 5.3"<<Message::done;

      case Y_MAX_DIR:
	return y_minor_derivative_edge_flow[from.pos];

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_flow : bad end 5.4"<<Message::done;

      case Z_MIN_x_max_DIR: 
	Message::error<<"edge_flow : bad end 5.5"<<Message::done;

      case Z_MIN_y_min_DIR:
	return -cost_edge_Z_MAX_y_max_flow[from.pos];

      case Z_MIN_y_max_DIR: 
	Message::error<<"edge_flow : bad end 5.6"<<Message::done;  

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_flow : bad end 5.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_flow : bad end 5.8"<<Message::done;

      case Z_MAX_y_min_DIR:
	return cost_edge_Z_MAX_MAX_y_max_flow[from.pos];

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_flow : bad end 5.9"<<Message::done;  
      }
    
      Message::error<<"edge_flow : bad end 5"<<Message::done;
      return 0;  
    }
  
    Message::error<<"edge_flow : bad end 0"<<Message::done;
    return 0;
  }

  
  template<typename Real>
  typename PushRelabel_non_linear_grid_3D<Real>::real_type
  PushRelabel_non_linear_grid_3D<Real>::
  edge_residual(const node_type&     from,
		const direction_type dir) const{

    switch(from.loc){
      // ########################### CENTER_NODE ###########################
    case CENTER_NODE:
    
      switch(dir){
      case X_MIN_DIR:
	return x_major_derivative_edge_capacity(from.pos.x()-1,from.pos.y())
	  + x_major_derivative_edge_flow(from.pos.x()-1,from.pos.y(),from.pos.z());
      
      case X_MAX_DIR:
	return x_major_derivative_edge_capacity(from.pos.x(),from.pos.y())
	  - x_major_derivative_edge_flow[from.pos];

      case Y_MIN_DIR:
	return y_major_derivative_edge_capacity(from.pos.x(),from.pos.y()-1)
	  + y_major_derivative_edge_flow(from.pos.x(),from.pos.y()-1,from.pos.z());
      
      case Y_MAX_DIR:
	return y_major_derivative_edge_capacity(from.pos.x(),from.pos.y())
	  - y_major_derivative_edge_flow[from.pos];

      case Z_MIN_x_min_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z()-1)
	  + cost_edge_Z_MAX_MAX_x_min_flow(from.pos.x(),from.pos.y(),from.pos.z()-1);
      
      case Z_MIN_x_max_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z()-1)
	  + cost_edge_Z_MAX_MAX_x_max_flow(from.pos.x(),from.pos.y(),from.pos.z()-1);

      case Z_MIN_y_min_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z()-1)
	  + cost_edge_Z_MAX_MAX_y_min_flow(from.pos.x(),from.pos.y(),from.pos.z()-1);
      
      case Z_MIN_y_max_DIR:
	return cost_edge_capacity(from.pos.x(),from.pos.y(),from.pos.z()-1)
	  + cost_edge_Z_MAX_MAX_y_max_flow(from.pos.x(),from.pos.y(),from.pos.z()-1);

      case Z_MAX_x_min_DIR:
	return cost_edge_capacity[from.pos]
	  - cost_edge_Z_MAX_x_min_flow[from.pos];
      
      case Z_MAX_x_max_DIR:
	return cost_edge_capacity[from.pos]
	  - cost_edge_Z_MAX_x_max_flow[from.pos];

      case Z_MAX_y_min_DIR:
	return cost_edge_capacity[from.pos]
	  - cost_edge_Z_MAX_y_min_flow[from.pos];
      
      case Z_MAX_y_max_DIR:
	return cost_edge_capacity[from.pos]
	  - cost_edge_Z_MAX_y_max_flow[from.pos];
      }

      Message::error<<"edge_residual : bad end 1"<<Message::done;
      return 0;
    
      // ########################### Z_MAX_x_min_NODE ###########################  
    case Z_MAX_x_min_NODE:

      switch(dir){
      case X_MIN_DIR:
	return x_minor_derivative_edge_capacity(from.pos.x()-1,from.pos.y())
	  + x_minor_derivative_edge_flow(from.pos.x()-1,from.pos.y(),from.pos.z());

      case X_MAX_DIR:
	Message::error<<"edge_residual : bad end 2.1"<<Message::done;

      case Y_MIN_DIR:
	Message::error<<"edge_residual : bad end 2.2"<<Message::done;

      case Y_MAX_DIR:
	Message::error<<"edge_residual : bad end 2.3"<<Message::done;  

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_residual : bad end 2.4"<<Message::done;
   
      case Z_MIN_x_max_DIR:
	return cost_edge_capacity[from.pos]
	  + cost_edge_Z_MAX_x_min_flow[from.pos];

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_residual : bad end 2.5"<<Message::done;
   
      case Z_MIN_y_max_DIR:
	Message::error<<"edge_residual : bad end 2.6"<<Message::done;

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_residual : bad end 2.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	return cost_edge_capacity[from.pos]
	  - cost_edge_Z_MAX_MAX_x_min_flow[from.pos];

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_residual : bad end 2.8"<<Message::done;

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_residual : bad end 2.9"<<Message::done;
      }
    
      Message::error<<"edge_residual : bad end 2"<<Message::done;
      return 0;
    
      // ########################### Z_MAX_x_max_NODE ###########################
    case Z_MAX_x_max_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_residual : bad end 3.1"<<Message::done;

      case X_MAX_DIR:
	return x_minor_derivative_edge_capacity(from.pos.x(),from.pos.y())
	  - x_minor_derivative_edge_flow[from.pos];

      case Y_MIN_DIR:
	Message::error<<"edge_residual : bad end 3.2"<<Message::done;

      case Y_MAX_DIR:
	Message::error<<"edge_residual : bad end 3.3"<<Message::done;

      case Z_MIN_x_min_DIR:
	return cost_edge_capacity[from.pos]
	  + cost_edge_Z_MAX_x_max_flow[from.pos];

      case Z_MIN_x_max_DIR: 
	Message::error<<"edge_residual : bad end 3.4"<<Message::done;

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_residual : bad end 3.5"<<Message::done;

      case Z_MIN_y_max_DIR: 
	Message::error<<"edge_residual : bad end 3.6"<<Message::done;  

      case Z_MAX_x_min_DIR:
	return cost_edge_capacity[from.pos]
	  - cost_edge_Z_MAX_MAX_x_max_flow[from.pos];

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_residual : bad end 3.7"<<Message::done;

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_residual : bad end 3.8"<<Message::done;

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_residual : bad end 3.9"<<Message::done;  
      }
    
      Message::error<<"edge_residual : bad end 3"<<Message::done;
      return 0;

      // ########################### Z_MAX_y_min_NODE ###########################  
    case Z_MAX_y_min_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_residual : bad end 4.1"<<Message::done;

      case X_MAX_DIR:
	Message::error<<"edge_residual : bad end 4.2"<<Message::done;

      case Y_MIN_DIR:
	return y_minor_derivative_edge_capacity(from.pos.x(),from.pos.y()-1)
	  + y_minor_derivative_edge_flow(from.pos.x(),from.pos.y()-1,from.pos.z());


      case Y_MAX_DIR:
	Message::error<<"edge_residual : bad end 4.3"<<Message::done;  

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_residual : bad end 4.4"<<Message::done;
   
      case Z_MIN_x_max_DIR:
	Message::error<<"edge_residual : bad end 4.5"<<Message::done;

      case Z_MIN_y_min_DIR:
	Message::error<<"edge_residual : bad end 4.6"<<Message::done;
   
      case Z_MIN_y_max_DIR:
	return cost_edge_capacity[from.pos]
	  + cost_edge_Z_MAX_y_min_flow[from.pos];

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_residual : bad end 4.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_residual : bad end 4.8"<<Message::done;

      case Z_MAX_y_min_DIR:
	Message::error<<"edge_residual : bad end 4.9"<<Message::done;

      case Z_MAX_y_max_DIR:
	return cost_edge_capacity[from.pos]
	  - cost_edge_Z_MAX_MAX_y_min_flow[from.pos];

      }
    
      Message::error<<"edge_residual : bad end 4"<<Message::done;
      return 0;
    
      // ########################### Z_MAX_y_max_NODE ###########################
    case Z_MAX_y_max_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"edge_residual : bad end 5.1"<<Message::done;

      case X_MAX_DIR:
	Message::error<<"edge_residual : bad end 5.2"<<Message::done;

      case Y_MIN_DIR:
	Message::error<<"edge_residual : bad end 5.3"<<Message::done;

      case Y_MAX_DIR:
	return y_minor_derivative_edge_capacity(from.pos.x(),from.pos.y())
	  - y_minor_derivative_edge_flow[from.pos];

      case Z_MIN_x_min_DIR:
	Message::error<<"edge_residual : bad end 5.4"<<Message::done;

      case Z_MIN_x_max_DIR: 
	Message::error<<"edge_residual : bad end 5.5"<<Message::done;

      case Z_MIN_y_min_DIR:
	return cost_edge_capacity[from.pos]
	  + cost_edge_Z_MAX_y_max_flow[from.pos];

      case Z_MIN_y_max_DIR: 
	Message::error<<"edge_residual : bad end 5.6"<<Message::done;  

      case Z_MAX_x_min_DIR:
	Message::error<<"edge_residual : bad end 5.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	Message::error<<"edge_residual : bad end 5.8"<<Message::done;

      case Z_MAX_y_min_DIR:
	return cost_edge_capacity[from.pos]
	  - cost_edge_Z_MAX_MAX_y_max_flow[from.pos];

      case Z_MAX_y_max_DIR:
	Message::error<<"edge_residual : bad end 5.9"<<Message::done;  
      }
    
      Message::error<<"edge_residual : bad end 5"<<Message::done;
      return 0;  
    }
  
    Message::error<<"edge_residual : bad end 0"<<Message::done;
    return 0;
  }



/*!
  No domain check.
*/
  template<typename Real>
  void
  PushRelabel_non_linear_grid_3D<Real>::
  neighbors(const node_type& node,
	    node_type* const result,
	    size_type* const n_res) const{
  
    *n_res = 0;

    // ################ toward X_MIN ################ 
  
    switch(node.loc){
      
    case CENTER_NODE:
      if ((node.pos.x()>0)
	  &&(domain(node.pos.x()-1,node.pos.y(),node.pos.z())||
	     ((node.pos.z()>0)&&
	      domain(node.pos.x()-1,node.pos.y(),node.pos.z()-1)))){
	result[*n_res].pos.x() = node.pos.x()-1;
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;
      
    case Z_MAX_x_min_NODE:
      if ((node.pos.x()>0)
	  &&(domain(node.pos.x()-1,node.pos.y(),node.pos.z()))){
	result[*n_res].pos.x() = node.pos.x()-1;
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = Z_MAX_x_max_NODE;
	(*n_res)++;
      }
      break;
      
    case Z_MAX_x_max_NODE:
    case Z_MAX_y_min_NODE:
    case Z_MAX_y_max_NODE:
      break;
    }

    // ################ toward Y_MIN ################ 
  
    switch(node.loc){
      
    case CENTER_NODE:
      if ((node.pos.y()>0)
	  &&(domain(node.pos.x(),node.pos.y()-1,node.pos.z())||
	     ((node.pos.z()>0)&&
	      domain(node.pos.x(),node.pos.y()-1,node.pos.z()-1)))){
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y()-1;
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;
      
    case Z_MAX_y_min_NODE:
      if ((node.pos.y()>0)
	  &&(domain(node.pos.x(),node.pos.y()-1,node.pos.z()))){
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y()-1;
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = Z_MAX_y_max_NODE;
	(*n_res)++;
      }
      break;
      
    case Z_MAX_x_max_NODE:
    case Z_MAX_x_min_NODE:
    case Z_MAX_y_max_NODE:
      break;
    }
  
    // ################ toward Z_MIN ################
  
    switch(node.loc){
      
    case CENTER_NODE:

      if ((node.pos.z()>0)
	  &&(domain(node.pos.x(),node.pos.y(),node.pos.z()-1))){

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()-1;
	result[*n_res].loc     = Z_MAX_x_min_NODE;
	(*n_res)++;

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()-1;
	result[*n_res].loc     = Z_MAX_x_max_NODE;
	(*n_res)++;

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()-1;
	result[*n_res].loc     = Z_MAX_y_min_NODE;
	(*n_res)++;

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()-1;
	result[*n_res].loc     = Z_MAX_y_max_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_x_min_NODE:
      result[*n_res].pos.x() = node.pos.x();
      result[*n_res].pos.y() = node.pos.y();
      result[*n_res].pos.z() = node.pos.z();
      result[*n_res].loc     = CENTER_NODE;
      (*n_res)++;
      break;

    case Z_MAX_x_max_NODE:
      result[*n_res].pos.x() = node.pos.x();
      result[*n_res].pos.y() = node.pos.y();
      result[*n_res].pos.z() = node.pos.z();
      result[*n_res].loc     = CENTER_NODE;
      (*n_res)++;
      break;

    case Z_MAX_y_min_NODE:
      result[*n_res].pos.x() = node.pos.x();
      result[*n_res].pos.y() = node.pos.y();
      result[*n_res].pos.z() = node.pos.z();
      result[*n_res].loc     = CENTER_NODE;
      (*n_res)++;
      break;

    case Z_MAX_y_max_NODE:
      result[*n_res].pos.x() = node.pos.x();
      result[*n_res].pos.y() = node.pos.y();
      result[*n_res].pos.z() = node.pos.z();
      result[*n_res].loc     = CENTER_NODE;
      (*n_res)++;
      break;  
    }

    // ################ toward X_MAX ################ 

    switch(node.loc){

    case CENTER_NODE:

      if ((node.pos.x()<x_size-1)
	  &&(domain(node.pos.x()+1,node.pos.y(),node.pos.z())||
	     ((node.pos.z()>0)&&
	      domain(node.pos.x()+1,node.pos.y(),node.pos.z()-1)))){

	result[*n_res].pos.x() = node.pos.x()+1;
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_x_max_NODE:

      if ((node.pos.x()<x_size-1)
	  &&(domain(node.pos.x()+1,node.pos.y(),node.pos.z()))){
	result[*n_res].pos.x() = node.pos.x()+1;
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = Z_MAX_x_min_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_x_min_NODE:
    case Z_MAX_y_min_NODE:
    case Z_MAX_y_max_NODE:
      break;  
    }

    // ################ toward Y_MAX ################ 

    switch(node.loc){

    case CENTER_NODE:

      if ((node.pos.y()<y_size-1)
	  &&(domain(node.pos.x(),node.pos.y()+1,node.pos.z())||
	     ((node.pos.z()>0)&&
	      domain(node.pos.x(),node.pos.y()+1,node.pos.z()-1)))){
    
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y()+1;
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_y_max_NODE:
      if ((node.pos.y()<y_size-1)
	  &&(domain(node.pos.x(),node.pos.y()+1,node.pos.z()))){

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y()+1;
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = Z_MAX_y_min_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_x_min_NODE:
    case Z_MAX_x_max_NODE:
    case Z_MAX_y_min_NODE:
      break;  
    }
    

    // ################ toward Z_MAX ################ 

    switch(node.loc){
    case CENTER_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])){

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = Z_MAX_x_min_NODE;
	(*n_res)++;
      
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = Z_MAX_x_max_NODE;
	(*n_res)++;
      
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = Z_MAX_y_min_NODE;
	(*n_res)++;
      
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = Z_MAX_y_max_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_x_min_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])){//no +1 because node ok if ZMIN ok.
      
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()+1;
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_x_max_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])){//no +1 because node ok if ZMIN ok.

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()+1;
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_y_min_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])){//no +1 because node ok if ZMIN ok.
	 
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()+1;
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_y_max_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])){//no +1 because node ok if ZMIN ok.
	
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()+1;
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;
    }
  }





/*!
  No domain check.
*/
  template<typename Real>
  void
  PushRelabel_non_linear_grid_3D<Real>::
  adjacent_nodes(const node_type& node,
		 node_type* const result,
		 size_type* const n_res) const{

    *n_res = 0;

    // ################ toward X_MIN ################ 
      
    switch(node.loc){
      
    case CENTER_NODE:
      if ((node.pos.x()>0)
	  &&(domain(node.pos.x()-1,node.pos.y(),node.pos.z())||
	     ((node.pos.z()>0)&&
	      domain(node.pos.x()-1,node.pos.y(),node.pos.z()-1)))){

	if(edge_residual(node,X_MIN_DIR)>epsilon){
	  result[*n_res].pos.x() = node.pos.x()-1;
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = CENTER_NODE;
	  (*n_res)++;
	}
      }
      break;
      
    case Z_MAX_x_min_NODE:
      if ((node.pos.x()>0)
	  &&(domain(node.pos.x()-1,node.pos.y(),node.pos.z()))){
	if(edge_residual(node,X_MIN_DIR)>epsilon){
	  result[*n_res].pos.x() = node.pos.x()-1;
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = Z_MAX_x_max_NODE;
	  (*n_res)++;
	}
      }
      break;
      
    case Z_MAX_x_max_NODE:
    case Z_MAX_y_min_NODE:
    case Z_MAX_y_max_NODE:
      break;
    }

    // ################ toward Y_MIN ################ 
  
    switch(node.loc){
      
    case CENTER_NODE:
      if ((node.pos.y()>0)
	  &&(domain(node.pos.x(),node.pos.y()-1,node.pos.z())||
	     ((node.pos.z()>0)&&
	      domain(node.pos.x(),node.pos.y()-1,node.pos.z()-1)))){
	if(edge_residual(node,Y_MIN_DIR)>epsilon){
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y()-1;
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = CENTER_NODE;
	  (*n_res)++;
	}
      }
      break;
      
    case Z_MAX_y_min_NODE:
      if ((node.pos.y()>0)
	  &&(domain(node.pos.x(),node.pos.y()-1,node.pos.z()))){
	if(edge_residual(node,Y_MIN_DIR)>epsilon){
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y()-1;
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = Z_MAX_y_max_NODE;
	  (*n_res)++;
	}
      }
      break;
      
    case Z_MAX_x_max_NODE:
    case Z_MAX_x_min_NODE:
    case Z_MAX_y_max_NODE:
      break;
    }

  
    // ################ toward Z_MIN ################
  
    switch(node.loc){
      
    case CENTER_NODE:

      if ((node.pos.z()>0)
	  &&(domain(node.pos.x(),node.pos.y(),node.pos.z()-1))){

	if (edge_residual(node,Z_MIN_x_min_DIR)>epsilon){
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z()-1;
	  result[*n_res].loc     = Z_MAX_x_min_NODE;
	  (*n_res)++;
	}

	if (edge_residual(node,Z_MIN_x_max_DIR)>epsilon){
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z()-1;
	  result[*n_res].loc     = Z_MAX_x_max_NODE;
	  (*n_res)++;
	}

	if (edge_residual(node,Z_MIN_y_min_DIR)>epsilon){
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z()-1;
	  result[*n_res].loc     = Z_MAX_y_min_NODE;
	  (*n_res)++;
	}

	if (edge_residual(node,Z_MIN_y_max_DIR)>epsilon){
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z()-1;
	  result[*n_res].loc     = Z_MAX_y_max_NODE;
	  (*n_res)++;
	}
      }
      break;

    case Z_MAX_x_min_NODE:
      if (edge_residual(node,Z_MIN_x_max_DIR)>epsilon){
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_x_max_NODE:
      if (edge_residual(node,Z_MIN_x_min_DIR)>epsilon){
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_y_min_NODE:
      if (edge_residual(node,Z_MIN_y_max_DIR)>epsilon){
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_y_max_NODE:
      if (edge_residual(node,Z_MIN_y_min_DIR)>epsilon){
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z();
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;  
    }

    // ################ toward X_MAX ################ 

    switch(node.loc){

    case CENTER_NODE:
      if ((node.pos.x()<x_size-1)
	  &&(domain(node.pos.x()+1,node.pos.y(),node.pos.z())||
	     ((node.pos.z()>0)&&
	      domain(node.pos.x()+1,node.pos.y(),node.pos.z()-1)))){
	if (edge_residual(node,X_MAX_DIR)>epsilon){
    
	  result[*n_res].pos.x() = node.pos.x()+1;
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = CENTER_NODE;
	  (*n_res)++;
	}
      }
      break;

    case Z_MAX_x_max_NODE:
      if ((node.pos.x()<x_size-1)
	  &&(domain(node.pos.x()+1,node.pos.y(),node.pos.z()))){
	if (edge_residual(node,X_MAX_DIR)>epsilon){
    
	  result[*n_res].pos.x() = node.pos.x()+1;
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = Z_MAX_x_min_NODE;
	  (*n_res)++;
	}
      }
      break;

    case Z_MAX_x_min_NODE:
    case Z_MAX_y_min_NODE:
    case Z_MAX_y_max_NODE:
      break;  
    }

    // ################ toward Y_MAX ################ 

    switch(node.loc){

    case CENTER_NODE:
      if ((node.pos.y()<y_size-1)
	  &&(domain(node.pos.x(),node.pos.y()+1,node.pos.z())||
	     ((node.pos.z()>0)&&
	      domain(node.pos.x(),node.pos.y()+1,node.pos.z()-1)))){

	if (edge_residual(node,Y_MAX_DIR)>epsilon){
    
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y()+1;
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = CENTER_NODE;
	  (*n_res)++;
	}
      }
      break;

    case Z_MAX_y_max_NODE:
      if ((node.pos.y()<y_size-1)
	  &&(domain(node.pos.x(),node.pos.y()+1,node.pos.z()))){
	if (edge_residual(node,Y_MAX_DIR)>epsilon){
    
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y()+1;
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = Z_MAX_y_min_NODE;
	  (*n_res)++;
	}
      }
      break;

    case Z_MAX_x_min_NODE:
    case Z_MAX_x_max_NODE:
    case Z_MAX_y_min_NODE:
      break;  
    }
  

    // ################ toward Z_MAX ################ 

    switch(node.loc){
    case CENTER_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])){

	if (edge_residual(node,Z_MAX_x_min_DIR)>epsilon){
      
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = Z_MAX_x_min_NODE;
	  (*n_res)++;
	}

	if (edge_residual(node,Z_MAX_x_max_DIR)>epsilon){
      
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = Z_MAX_x_max_NODE;
	  (*n_res)++;
	}

	if (edge_residual(node,Z_MAX_y_min_DIR)>epsilon){
      
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = Z_MAX_y_min_NODE;
	  (*n_res)++;
	}

	if (edge_residual(node,Z_MAX_y_max_DIR)>epsilon){
      
	  result[*n_res].pos.x() = node.pos.x();
	  result[*n_res].pos.y() = node.pos.y();
	  result[*n_res].pos.z() = node.pos.z();
	  result[*n_res].loc     = Z_MAX_y_max_NODE;
	  (*n_res)++;
	}
      }
      break;

    case Z_MAX_x_min_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])//no +1 because node ok if ZMIN ok.
	  &&(edge_residual(node,Z_MAX_x_max_DIR)>epsilon)){
      
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()+1;
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_x_max_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])//no +1 because node ok if ZMIN ok.
	  &&(edge_residual(node,Z_MAX_x_min_DIR)>epsilon)){

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()+1;
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_y_min_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])//no +1 because node ok if ZMIN ok.
	  &&(edge_residual(node,Z_MAX_y_max_DIR)>epsilon)){
      
	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()+1;
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;

    case Z_MAX_y_max_NODE:
      if ((node.pos.z()<z_size)
	  &&(domain[node.pos])//no +1 because node ok if ZMIN ok.
	  &&(edge_residual(node,Z_MAX_y_min_DIR)>epsilon)){

	result[*n_res].pos.x() = node.pos.x();
	result[*n_res].pos.y() = node.pos.y();
	result[*n_res].pos.z() = node.pos.z()+1;
	result[*n_res].loc     = CENTER_NODE;
	(*n_res)++;
      }
      break;
    }
  }


  template<typename Real>
  typename PushRelabel_non_linear_grid_3D<Real>::real_type
  PushRelabel_non_linear_grid_3D<Real>::
  node_excess(const node_type& node) const{
  
    // neigbors of the source have an infinite excess.
    if ((node.loc == CENTER_NODE) && (node.pos.z() == min_bound(node.pos.x(),
								node.pos.y()))){
#ifndef WITHOUT_LIMITS
      return (std::numeric_limits<real_type>::max());
#else
      return 1e20;
#endif
    }
  
    // neighbors of the think have no excess.
    if ((node.loc == CENTER_NODE) && (node.pos.z() == max_bound(node.pos.x(),
								node.pos.y()))){
      return 0;
    } 
  
    node_type neighbor[12];
    size_type n_neighbor;

    neighbors(node,neighbor,&n_neighbor);

    real_type res = 0;
    for(size_type n=0;n<n_neighbor;n++){
      res -= edge_flow(node,direction(node,neighbor[n]));
    }
    return res;
  }


/*!
  No domain check.
*/
  template<typename Real>
  void
  PushRelabel_non_linear_grid_3D<Real>::
  add_flow(const node_type&     from,
	   const direction_type dir,
	   const real_type      flow){

//     if (edge_residual(from,dir)<=flow){
    if (edge_residual(from,dir)+epsilon<flow){
      Message::error<<"not enough capacity ("<<edge_residual(from,dir)
		    <<") for flow ("<<flow<<") ["
		    <<(flow-edge_residual(from,dir))<<"]"
		    <<Message::done;
    }

    switch(from.loc){
      // ########################### CENTER_NODE ###########################
    case CENTER_NODE:
    
      switch(dir){
      case X_MIN_DIR:
	x_major_derivative_edge_flow(from.pos.x()-1,from.pos.y(),from.pos.z())
	  -= flow;
	return;
      
      case X_MAX_DIR:
	x_major_derivative_edge_flow[from.pos]
	  += flow;
	return;

      case Y_MIN_DIR:
	y_major_derivative_edge_flow(from.pos.x(),from.pos.y()-1,from.pos.z())
	  -= flow;
	return;
      
      case Y_MAX_DIR:
	y_major_derivative_edge_flow[from.pos]
	  += flow;
	return;  

      case Z_MIN_x_min_DIR:
	cost_edge_Z_MAX_MAX_x_min_flow(from.pos.x(),from.pos.y(),from.pos.z()-1)
	  -= flow;
	return;
      
      case Z_MIN_x_max_DIR:
	cost_edge_Z_MAX_MAX_x_max_flow(from.pos.x(),from.pos.y(),from.pos.z()-1)
	  -= flow;
	return;

      case Z_MIN_y_min_DIR:
	cost_edge_Z_MAX_MAX_y_min_flow(from.pos.x(),from.pos.y(),from.pos.z()-1)
	  -= flow;
	return;
      
      case Z_MIN_y_max_DIR:
	cost_edge_Z_MAX_MAX_y_max_flow(from.pos.x(),from.pos.y(),from.pos.z()-1)
	  -= flow;
	return;  

      case Z_MAX_x_min_DIR:
	cost_edge_Z_MAX_x_min_flow[from.pos]
	  += flow;
	return;
     
      case Z_MAX_x_max_DIR:
	cost_edge_Z_MAX_x_max_flow[from.pos]
	  += flow;
	return;

      case Z_MAX_y_min_DIR:
	cost_edge_Z_MAX_y_min_flow[from.pos]
	  += flow;
	return;
     
      case Z_MAX_y_max_DIR:
	cost_edge_Z_MAX_y_max_flow[from.pos]
	  += flow;
	return;  
      }

      Message::error<<"add_flow: bad end 1"<<Message::done;
    
      // ########################### Z_MAX_x_min_NODE ###########################  
    case Z_MAX_x_min_NODE:

      switch(dir){
      case X_MIN_DIR:
	x_minor_derivative_edge_flow(from.pos.x()-1,from.pos.y(),from.pos.z())
	  -= flow;
	return;

      case X_MAX_DIR:
	Message::error<<"add_flow : bad end 2.1"<<Message::done;

      case Y_MIN_DIR:
	Message::error<<"add_flow : bad end 2.2"<<Message::done;
      
      case Y_MAX_DIR:
	Message::error<<"add_flow : bad end 2.3"<<Message::done;

      case Z_MIN_x_min_DIR:
	Message::error<<"add_flow : bad end 2.4"<<Message::done;
   
      case Z_MIN_x_max_DIR:
	cost_edge_Z_MAX_x_min_flow[from.pos]
	  -= flow;
	return;

      case Z_MIN_y_min_DIR:
	Message::error<<"add_flow : bad end 2.5"<<Message::done;
   
      case Z_MIN_y_max_DIR:
	Message::error<<"add_flow : bad end 2.6"<<Message::done;

      case Z_MAX_x_min_DIR:
	Message::error<<"add_flow : bad end 2.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	cost_edge_Z_MAX_MAX_x_min_flow[from.pos]
	  += flow;
	return;

      case Z_MAX_y_min_DIR:
	Message::error<<"add_flow : bad end 2.8"<<Message::done;

      case Z_MAX_y_max_DIR:
	Message::error<<"add_flow : bad end 2.9"<<Message::done;
      }
    
      Message::error<<"add_flow : bad end 2"<<Message::done;
    
      // ########################### Z_MAX_x_max_NODE ###########################
    case Z_MAX_x_max_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"add_flow : bad end 3.1"<<Message::done;

      case X_MAX_DIR:
	x_minor_derivative_edge_flow[from.pos]
	  += flow;
	return;

      case Y_MIN_DIR:
	Message::error<<"add_flow : bad end 3.2"<<Message::done;

      case Y_MAX_DIR:
	Message::error<<"add_flow : bad end 3.3"<<Message::done;

      case Z_MIN_x_min_DIR:
	cost_edge_Z_MAX_x_max_flow[from.pos]
	  -= flow;
	return;

      case Z_MIN_x_max_DIR: 
	Message::error<<"add_flow : bad end 3.4"<<Message::done;

      case Z_MIN_y_min_DIR:
	Message::error<<"add_flow : bad end 3.5"<<Message::done;

      case Z_MIN_y_max_DIR: 
	Message::error<<"add_flow : bad end 3.6"<<Message::done;  

      case Z_MAX_x_min_DIR:
	cost_edge_Z_MAX_MAX_x_max_flow[from.pos]
	  += flow;
	return;

      case Z_MAX_x_max_DIR:
	Message::error<<"add_flow : bad end 3.7"<<Message::done;

      case Z_MAX_y_min_DIR:
	Message::error<<"add_flow : bad end 3.8"<<Message::done;

      case Z_MAX_y_max_DIR:
	Message::error<<"add_flow : bad end 3.9"<<Message::done;  
      }
    
      Message::error<<"add_flow : bad end 3"<<Message::done;

      // ########################### Z_MAX_y_min_NODE ###########################  
    case Z_MAX_y_min_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"add_flow : bad end 4.1"<<Message::done;

      case X_MAX_DIR:
	Message::error<<"add_flow : bad end 4.2"<<Message::done;

      case Y_MIN_DIR:
	y_minor_derivative_edge_flow(from.pos.x(),from.pos.y()-1,from.pos.z())
	  -= flow;
	return; 
      
      case Y_MAX_DIR:
	Message::error<<"add_flow : bad end 4.3"<<Message::done;

      case Z_MIN_x_min_DIR:
	Message::error<<"add_flow : bad end 4.4"<<Message::done;
   
      case Z_MIN_x_max_DIR:
	Message::error<<"add_flow : bad end 4.5"<<Message::done;

      case Z_MIN_y_min_DIR:
	Message::error<<"add_flow : bad end 4.6"<<Message::done;
   
      case Z_MIN_y_max_DIR:
	cost_edge_Z_MAX_y_min_flow[from.pos]
	  -= flow;
	return;

      case Z_MAX_x_min_DIR:
	Message::error<<"add_flow : bad end 4.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	Message::error<<"add_flow : bad end 4.8"<<Message::done;

      case Z_MAX_y_min_DIR:
	Message::error<<"add_flow : bad end 4.9"<<Message::done;

      case Z_MAX_y_max_DIR:
	cost_edge_Z_MAX_MAX_y_min_flow[from.pos]
	  += flow;
	return;
      }
    
      Message::error<<"add_flow : bad end 4"<<Message::done;
    
      // ########################### Z_MAX_y_max_NODE ###########################
    case Z_MAX_y_max_NODE:

      switch(dir){
      case X_MIN_DIR:
	Message::error<<"add_flow : bad end 5.1"<<Message::done;

      case X_MAX_DIR:
	Message::error<<"add_flow : bad end 5.2"<<Message::done;

      case Y_MIN_DIR:
	Message::error<<"add_flow : bad end 5.3"<<Message::done;

      case Y_MAX_DIR:
	y_minor_derivative_edge_flow[from.pos]
	  += flow;
	return;

      case Z_MIN_x_min_DIR:
	Message::error<<"add_flow : bad end 5.4"<<Message::done;

      case Z_MIN_x_max_DIR: 
	Message::error<<"add_flow : bad end 5.5"<<Message::done;

      case Z_MIN_y_min_DIR:
	cost_edge_Z_MAX_y_max_flow[from.pos]
	  -= flow;
	return;

      case Z_MIN_y_max_DIR: 
	Message::error<<"add_flow : bad end 5.6"<<Message::done;  

      case Z_MAX_x_min_DIR:
	Message::error<<"add_flow : bad end 5.7"<<Message::done;

      case Z_MAX_x_max_DIR:
	Message::error<<"add_flow : bad end 5.8"<<Message::done;

      case Z_MAX_y_min_DIR:
	cost_edge_Z_MAX_MAX_y_max_flow[from.pos]
	  += flow;
	return;

      case Z_MAX_y_max_DIR:
	Message::error<<"add_flow : bad end 5.9"<<Message::done;  
      }
    
      Message::error<<"add_flow : bad end 5"<<Message::done;
    }
  
    Message::error<<"add_flow : bad end 0"<<Message::done;
  }


  template<typename Real>
  typename PushRelabel_non_linear_grid_3D<Real>::direction_type
  PushRelabel_non_linear_grid_3D<Real>::
  direction(const node_type& from,
	    const node_type& to) const{

    switch(from.loc){

      // ########################### CENTER_NODE ###########################
    case CENTER_NODE:
      switch(to.loc){
      
      case CENTER_NODE:
	if (from.pos.z() == to.pos.z()){
      
	  if ((from.pos.x()+1 == to.pos.x()) && (from.pos.y() == to.pos.y())){
	    return X_MAX_DIR;
	  }
	  else if ((from.pos.x() == to.pos.x()+1) && (from.pos.y() == to.pos.y())){
	    return X_MIN_DIR;
	  }
	  else if ((from.pos.x() == to.pos.x()) && (from.pos.y()+1 == to.pos.y())){
	    return Y_MAX_DIR;
	  }
	  else if ((from.pos.x() == to.pos.x()) && (from.pos.y() == to.pos.y()+1)){
	    return Y_MIN_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 1.1"<<Message::done;
	    return 0;
	  }
	}
	else{
	  Message::error<<"direction : bad end 1.2"<<Message::done;
	  return 0;
	}
          
      case Z_MAX_x_min_NODE:
	if ((from.pos.x() == to.pos.x()) && (from.pos.y() == to.pos.y())){
	  if (to.pos.z()+1 == from.pos.z()){
	    return Z_MIN_x_min_DIR;
	  }
	  else if (from.pos.z() == to.pos.z()){
	    return Z_MAX_x_min_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 1.3"<<Message::done;
	    return 0;
	  }
	}
	else{
	  Message::error<<"direction : bad end 1.4"<<Message::done;
	  return 0;
	}
    
      case Z_MAX_x_max_NODE:
	if ((from.pos.x() == to.pos.x()) && (from.pos.y() == to.pos.y())){
	  if (to.pos.z()+1 == from.pos.z()){
	    return Z_MIN_x_max_DIR;
	  }
	  else if (from.pos.z() == to.pos.z()){
	    return Z_MAX_x_max_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 1.5"<<Message::done;
	    return 0;
	  }
	}
	else{
	  Message::error<<"direction : bad end 1.6"<<Message::done;
	  return 0;
	}

      case Z_MAX_y_min_NODE:
	if ((from.pos.x() == to.pos.x()) && (from.pos.y() == to.pos.y())){
	  if (to.pos.z()+1 == from.pos.z()){
	    return Z_MIN_y_min_DIR;
	  }
	  else if (from.pos.z() == to.pos.z()){
	    return Z_MAX_y_min_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 1.7"<<Message::done;
	    return 0;
	  }
	}
	else{
	  Message::error<<"direction : bad end 1.8\n"
			<<VALUE_OF(from.pos)<<"\t"<<VALUE_OF(int(from.loc))<<"\n"
			<<VALUE_OF(to.pos)<<"\t"<<VALUE_OF(int(to.loc))
			<<Message::done;
	  return 0;
	}
    
      case Z_MAX_y_max_NODE:
	if ((from.pos.x() == to.pos.x()) && (from.pos.y() == to.pos.y())){
	  if (to.pos.z()+1 == from.pos.z()){
	    return Z_MIN_y_max_DIR;
	  }
	  else if (from.pos.z() == to.pos.z()){
	    return Z_MAX_y_max_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 1.9"<<Message::done;
	    return 0;
	  }
	}
	else{
	  Message::error<<"direction : bad end 1.10"<<Message::done;
	  return 0;
	}  
      }

      // ########################### Z_MAX_x_min_NODE ###########################  
    case Z_MAX_x_min_NODE:
      switch(to.loc){
      
      case CENTER_NODE:
	if ((from.pos.x() == to.pos.x()) && (from.pos.y() == to.pos.y())){
	  if (from.pos.z() == to.pos.z()){
	    return Z_MIN_x_max_DIR;
	  }
	  else if (from.pos.z()+1 == to.pos.z()){
	    return Z_MAX_x_max_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 2.1"<<Message::done;
	    return 0;
	  }
	}
      
      case Z_MAX_x_min_NODE:
	Message::error<<"direction : bad end 2.2"<<Message::done;
	return 0;
      
    
      case Z_MAX_x_max_NODE:
	if ((from.pos.y() == to.pos.y())&&
	    (from.pos.z() == to.pos.z())&&
	    (from.pos.x() == to.pos.x()+1)){
	  return X_MIN_DIR;
	}
	else{
	  Message::error<<"direction : bad end 2.3 "<<Message::done;
	  return 0;
	}

      case Z_MAX_y_min_NODE:
	Message::error<<"direction : bad end 2.4"<<Message::done;
	return 0;
      
    
      case Z_MAX_y_max_NODE:
	Message::error<<"direction : bad end 2.5 "<<Message::done;
	return 0;
      }
    
      // ########################### Z_MAX_x_max_NODE ########################### 
    case Z_MAX_x_max_NODE:
      switch(to.loc){
      
      case CENTER_NODE:
	if ((from.pos.x() == to.pos.x())&&(from.pos.y() == to.pos.y())){
	  if (from.pos.z() == to.pos.z()){
	    return Z_MIN_x_min_DIR;
	  }
	  else if (from.pos.z()+1 == to.pos.z()){
	    return Z_MAX_x_min_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 3.1"<<Message::done;
	    return 0;
	  }
	}

      case Z_MAX_x_min_NODE:
	if ((from.pos.y() == to.pos.y())&&
	    (from.pos.z() == to.pos.z())&&
	    (from.pos.x()+1 == to.pos.x())){
	  return X_MAX_DIR;
	}      
    
      case Z_MAX_x_max_NODE:
	Message::error<<"direction : bad end 3.2"<<Message::done;
	return 0;

      case Z_MAX_y_min_NODE:
	Message::error<<"direction : bad end 3.3"<<Message::done;
	return 0;
      
      case Z_MAX_y_max_NODE:
	Message::error<<"direction : bad end 3.4"<<Message::done;
	return 0;
      }

      // ########################### Z_MAX_y_min_NODE ###########################  
    case Z_MAX_y_min_NODE:
      switch(to.loc){
      
      case CENTER_NODE:
	if ((from.pos.x() == to.pos.x()) && (from.pos.y() == to.pos.y())){
	  if (from.pos.z() == to.pos.z()){
	    return Z_MIN_y_max_DIR;
	  }
	  else if (from.pos.z()+1 == to.pos.z()){
	    return Z_MAX_y_max_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 4.1"<<Message::done;
	    return 0;
	  }
	}
      
      case Z_MAX_x_min_NODE:
	Message::error<<"direction : bad end 4.2"<<Message::done;
	return 0;
      
    
      case Z_MAX_x_max_NODE:
	Message::error<<"direction : bad end 4.3 "<<Message::done;
	return 0;      

      case Z_MAX_y_min_NODE:
	Message::error<<"direction : bad end 4.4"<<Message::done;
	return 0;
      
    
      case Z_MAX_y_max_NODE:
	if ((from.pos.x() == to.pos.x())&&
	    (from.pos.z() == to.pos.z())&&
	    (from.pos.y() == to.pos.y()+1)){
	  return Y_MIN_DIR;
	}
	else{
	  Message::error<<"direction : bad end 4.5 "<<Message::done;
	  return 0;
	}
      }
    
      // ########################### Z_MAX_y_max_NODE ########################### 
    case Z_MAX_y_max_NODE:
      switch(to.loc){
      
      case CENTER_NODE:
	if ((from.pos.x() == to.pos.x())&&(from.pos.y() == to.pos.y())){
	  if (from.pos.z() == to.pos.z()){
	    return Z_MIN_y_min_DIR;
	  }
	  else if (from.pos.z()+1 == to.pos.z()){
	    return Z_MAX_y_min_DIR;
	  }
	  else{
	    Message::error<<"direction : bad end 5.1"<<Message::done;
	    return 0;
	  }
	}

      case Z_MAX_x_min_NODE:    
	Message::error<<"direction : bad end 5.2"<<Message::done;
	return 0;
      
      case Z_MAX_x_max_NODE:
	Message::error<<"direction : bad end 5.3"<<Message::done;
	return 0;

      case Z_MAX_y_min_NODE:
	if ((from.pos.x() == to.pos.x())&&
	    (from.pos.z() == to.pos.z())&&
	    (from.pos.y()+1 == to.pos.y())){
	  return Y_MAX_DIR;
	}      
      
      case Z_MAX_y_max_NODE:
	Message::error<<"direction : bad end 5.4"<<Message::done;
	return 0;
      }
  
    }

    Message::error<<"direction : bad end 0"<<Message::done;
    return 0;
  }


  template<typename Real>
  typename PushRelabel_non_linear_grid_3D<Real>::label_type
  PushRelabel_non_linear_grid_3D<Real>::
  relabel(const node_type& node){
  
    node_type neighbor[12];
    size_type n_neighbor;
#ifndef WITHOUT_LIMITS
    label_type min = std::numeric_limits<label_type>::max();
#else
    label_type min = static_cast<size_type>(4294967295);
#endif

    static unsigned int NB_RELABEL = 0;
    static unsigned int NB_GLOBAL = 0;
  
    if (NB_RELABEL >= nb_nodes/2 /*10*/){
      global_relabel();
      NB_GLOBAL++;
      NB_RELABEL = 0;
      return 0;
    }


    NB_RELABEL++;

    // neighbors of the source have the source label.
    if ((node.pos.z() == min_bound(node.pos.x(),node.pos.y())) &&
	(node.loc == CENTER_NODE)){
    
      label_CENTER[node.pos] = nb_nodes;

      return nb_nodes;
    }
  
    adjacent_nodes(node,neighbor,&n_neighbor);

    for(size_type n=0;n<n_neighbor;n++){
      const label_type l = label(neighbor[n]);

      if (l<min){
	min = l;
      }
    }// end of for i

#ifndef WITHOUT_LIMITS
    if (min == std::numeric_limits<label_type>::max()){
#else
    if (min == static_cast<size_type>(4294967295)){
#endif
      Message::error<<"#neighbor = "<<n_neighbor<<std::endl;
      Message::error<<"labels :"<<std::endl;
      for(size_type n=0;n<n_neighbor;n++){
	const label_type l = label(neighbor[n]);

	Message::error<<l<<std::endl;
      }// end of for i

      Message::error<<"relabel : no neighbor reachable !"<<Message::done;
    }
  
    min++;

    label(node) = min;
  
    return min;
  }


  namespace{
    template<typename Real>
    struct Index_cmp{
      typedef typename PushRelabel_non_linear_grid_3D<Real>::index_type
      index_type;

      typedef typename PushRelabel_non_linear_grid_3D<Real>::node_type
      node_type;
    
      const PushRelabel_non_linear_grid_3D<Real>& graph;
     
      Index_cmp(const PushRelabel_non_linear_grid_3D<Real>& g):graph(g){}

      inline bool operator()(const index_type i1,
			     const index_type i2){
	node_type n1,n2;

	graph.from_index(&n1,i1);
	graph.from_index(&n2,i2);

	return (graph.label(n1)>graph.label(n2));
      }
    };
  }


  template<typename Real>
  void
  PushRelabel_non_linear_grid_3D<Real>::
  global_relabel(){

    using namespace std;

    cout<<"GLOBAL RELABEL ";

    deque<node_type>* to_explore = new deque<node_type>;
    deque<node_type>* exploring  = new deque<node_type>;
    typedef typename deque<node_type>::iterator iterator;

    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
	for(size_type z=0;z<z_size+1;z++){
      
	  label_CENTER(x,y,z) = 0;
	  label_Z_MAX_x_min(x,y,z) = 0;
	  label_Z_MAX_x_max(x,y,z) = 0;
	  label_Z_MAX_y_min(x,y,z) = 0;
	  label_Z_MAX_y_max(x,y,z) = 0;
	}
      }
    }

    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
	const size_type z = max_bound(x,y);
	node_type node(position_type(x,y,z),CENTER_NODE);
      
	if ((domain(x,y,z-1)) && (label_CENTER(x,y,z) < nb_nodes)){
	
	  exploring->push_back(node);
	  label_CENTER(x,y,z) = 1;
	}
      }
    }

    while(!exploring->empty()){
      for(iterator node=exploring->begin();node!=exploring->end();node++){
      
	node_type neighbor[12];
	size_type n_neighbor;

	neighbors(*node,neighbor,&n_neighbor);

	const label_type l = label(*node);
      
	for(size_type n=0;n<n_neighbor;n++){
	  const node_type n_node   = neighbor[n];
	  const direction_type dir = direction(n_node,*node);

	  if (edge_residual(n_node,dir)>epsilon){
	
	    const label_type ln = label(n_node);
	  
	    if (((ln > l+1)||(ln == 0))&&(ln < nb_nodes)){

	      label(n_node) = l+1;
	    
	      to_explore->push_back(n_node);

	    }
	  }
	
	}// end of for n
      }// end of for node

      delete exploring;
      exploring = to_explore;
      to_explore = new deque<node_type>;
    }// end of while
  
    delete exploring;
    delete to_explore;

    label_type max_l = 0;
    label_type max_L = 0;
    active_node_index.clear();

    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
	for(size_type z=0;z<z_size+1;z++){
      
	  if (label_CENTER(x,y,z) == 0) {
	    label_CENTER(x,y,z) = nb_nodes;
	  }

	  if (label_Z_MAX_x_min(x,y,z) == 0) {
	    label_Z_MAX_x_min(x,y,z) = nb_nodes;
	  }

	  if (label_Z_MAX_x_max(x,y,z) == 0) {
	    label_Z_MAX_x_max(x,y,z) = nb_nodes;
	  }

	  if (label_Z_MAX_y_min(x,y,z) == 0) {
	    label_Z_MAX_y_min(x,y,z) = nb_nodes;
	  }

	  if (label_Z_MAX_y_max(x,y,z) == 0) {
	    label_Z_MAX_y_max(x,y,z) = nb_nodes;
	  }

	  ///////
	  label_type l = label_CENTER(x,y,z);
	  node_type n = node_type(position_type(x,y,z),CENTER_NODE);
	  if ((l>max_L)&&(l<nb_nodes)){
	    max_L = l;
	  }
	  if ((l<nb_nodes)&&(node_excess(n)>epsilon)){
	    if (l>max_l){
	      max_l = l;
	    }
	  
	    index_type i;
	    to_index(n,&i);
	    active_node_index.push_back(i);
	  }

	  l = label_Z_MAX_x_min(x,y,z);
	  n = node_type(position_type(x,y,z),Z_MAX_x_min_NODE);
	  if ((l>max_L)&&(l<nb_nodes)){
	    max_L = l;
	  }
	  if ((l<nb_nodes)&&(node_excess(n)>epsilon)){
	    if (l>max_l){
	      max_l = l;
	    }

	    index_type i;
	    to_index(n,&i);
	    active_node_index.push_back(i);
	  }

	  l = label_Z_MAX_x_max(x,y,z);
	  n = node_type(position_type(x,y,z),Z_MAX_x_max_NODE);
	  if ((l>max_L)&&(l<nb_nodes)){
	    max_L = l;
	  }
	  if ((l<nb_nodes)&&(node_excess(n)>epsilon)){
	    if (l>max_l){
	      max_l = l;
	    }

	    index_type i;
	    to_index(n,&i);
	    active_node_index.push_back(i);
	  }

	  l = label_Z_MAX_y_min(x,y,z);
	  n = node_type(position_type(x,y,z),Z_MAX_y_min_NODE);
	  if ((l>max_L)&&(l<nb_nodes)){
	    max_L = l;
	  }
	  if ((l<nb_nodes)&&(node_excess(n)>epsilon)){
	    if (l>max_l){
	      max_l = l;
	    }

	    index_type i;
	    to_index(n,&i);
	    active_node_index.push_back(i);
	  }

	  l = label_Z_MAX_y_max(x,y,z);
	  n = node_type(position_type(x,y,z),Z_MAX_y_max_NODE);
	  if ((l>max_L)&&(l<nb_nodes)){
	    max_L = l;
	  }
	  if ((l<nb_nodes)&&(node_excess(n)>epsilon)){
	    if (l>max_l){
	      max_l = l;
	    }

	    index_type i;
	    to_index(n,&i);
	    active_node_index.push_back(i);
	  }
	  //////
	}
      }
    }

    Index_cmp<real_type> cmp(*this);
  
    std::sort(active_node_index.begin(),active_node_index.end(),cmp);
//    active_node_index.sort(cmp);

    cout<<"label max : "<<max_l<<" "<<max_L<<endl;
  
  }


/*!
  \e true if the operation have been done.
*/
  template<typename Real>
  bool
  PushRelabel_non_linear_grid_3D<Real>::
  discharge(const node_type& node){

    using namespace std;
    
    node_type neighbor[12];
    size_type n_neighbor;
    
    adjacent_nodes(node,neighbor,&n_neighbor);

    real_type excess = node_excess(node);
    size_type n = 0;

    if (excess <= epsilon) {
      return false;
    }

  
    while ((n<n_neighbor)&&(excess>epsilon)){

      const node_type n_node = neighbor[n];

      if (is_admissible(node,n_node)){
      
	const direction_type d = direction(node,n_node);
	const real_type to_push = std::min(excess,edge_residual(node,d));
	
	add_flow(node,d,to_push);

	excess -= to_push;
      
	add_active_node(n_node);
      }
      n++;    
    }

    if ((n==n_neighbor)&&(excess>epsilon)){
      relabel(node);
      add_active_node(node);
    }

    return true;
  }



  template<typename Real>
  void
  PushRelabel_non_linear_grid_3D<Real>::
  add_active_node(const node_type& node){

    index_type i;

    to_index(node,&i);

    active_node_index.push_back(i);
  }



  template<typename Real>
  void
  PushRelabel_non_linear_grid_3D<Real>::
  get_next_node(node_type* const node,
		bool* const      no_more_node){

    label_type l;
  
    do{
      *no_more_node = active_node_index.empty();

      if(!(*no_more_node)){
    
	const index_type i = active_node_index.front();
	active_node_index.pop_front();
      
	from_index(node,i);
        
	l = label(*node);
      }
    
    } while(((l >= nb_nodes)||
	     (node->pos.z() == max_bound(node->pos.x(),node->pos.y()))) &&
	    (!(*no_more_node)));
  }



  template<typename Real>
  bool
  PushRelabel_non_linear_grid_3D<Real>::
  is_active(const node_type& node) const{

    return (node_excess(node) > epsilon);
  }



  template<typename Real>
  bool
  PushRelabel_non_linear_grid_3D<Real>::
  is_admissible(const node_type& from,
		const node_type& to) const{

    return (label(from) == (label(to)+1));
  }


  template<typename Real>
  typename PushRelabel_non_linear_grid_3D<Real>::real_type
  PushRelabel_non_linear_grid_3D<Real>::
  max_flow(){

    using namespace std;

    global_relabel();

    node_type to_discharge;
    bool no_more_to_discharge;

    get_next_node(&to_discharge,&no_more_to_discharge);

    while(!no_more_to_discharge){
      discharge(to_discharge);
      get_next_node(&to_discharge,&no_more_to_discharge);
    }
  
    // Computes the flow.
    real_type flow = 0;
    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
    
	const size_type z = max_bound(x,y) - 1;
      
	if (domain(x,y,z)){
	  flow += cost_edge_Z_MAX_x_min_flow(x,y,z)
	    + cost_edge_Z_MAX_x_max_flow(x,y,z)
	    + cost_edge_Z_MAX_y_min_flow(x,y,z)
	    + cost_edge_Z_MAX_y_max_flow(x,y,z);
	}
      }
    }
  
    return flow;
  }



  template<typename Real>
  void
  PushRelabel_non_linear_grid_3D<Real>::
  min_cut(Array_2D<size_type>* const cut) const{
  
    using namespace std;
  
    Array_3D<bool> visited_CENTER(x_size,y_size,z_size+1,false);
    Array_3D<bool> visited_Z_MAX_x_min(x_size,y_size,z_size+1,false);
    Array_3D<bool> visited_Z_MAX_x_max(x_size,y_size,z_size+1,false);
    Array_3D<bool> visited_Z_MAX_y_min(x_size,y_size,z_size+1,false);
    Array_3D<bool> visited_Z_MAX_y_max(x_size,y_size,z_size+1,false);
  
    list<node_type> to_visit;

    // Initialization.
    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){
	node_type p(position_type(x,y,max_bound(x,y)),CENTER_NODE);    
	if (domain(p.pos.x(),p.pos.y(),p.pos.z()-1)&&
	    (label_CENTER[p.pos] < nb_nodes)){
	  to_visit.push_back(p);
	  visited_CENTER[p.pos] = true;
	}
      }
    }
  
    // Visits the graph to mark the nodes which are on the source side.
    node_type neighbor[12];
    size_type n_neighbor;
  
    while(!to_visit.empty()){
      node_type node = to_visit.front();
      to_visit.pop_front();

      neighbors(node,neighbor,&n_neighbor);
    
      for(size_type n=0;n<n_neighbor;n++){

	const node_type n_node   = neighbor[n];
	const direction_type dir = direction(n_node,node);
      
	if ((edge_residual(n_node,dir)>epsilon)/*&&(label(n_node)<nb_nodes)*/){
	
	  switch(n_node.loc){
	  case CENTER_NODE:
	  
	    if(!visited_CENTER[neighbor[n].pos]){
	      visited_CENTER[neighbor[n].pos] = true;
	      to_visit.push_back(neighbor[n]);
	    }
	    break;
	  
	  case Z_MAX_x_min_NODE:
	  
	    if(!visited_Z_MAX_x_min[neighbor[n].pos]){
	      visited_Z_MAX_x_min[neighbor[n].pos] = true;
	      to_visit.push_back(neighbor[n]);
	    }
	    break;
	  
	  case Z_MAX_x_max_NODE:
	  
	    if(!visited_Z_MAX_x_max[neighbor[n].pos]){
	      visited_Z_MAX_x_max[neighbor[n].pos] = true;
	      to_visit.push_back(neighbor[n]);
	    }
	    break;
	  
	  case Z_MAX_y_min_NODE:
	  
	    if(!visited_Z_MAX_y_min[neighbor[n].pos]){
	      visited_Z_MAX_y_min[neighbor[n].pos] = true;
	      to_visit.push_back(neighbor[n]);
	    }
	    break;
	  
	  case Z_MAX_y_max_NODE:
	  
	    if(!visited_Z_MAX_y_max[neighbor[n].pos]){
	      visited_Z_MAX_y_max[neighbor[n].pos] = true;
	      to_visit.push_back(neighbor[n]);
	    }
	    break;
	  }
	}
      }
    }

    // Looks for the cut
    cut->assign(x_size,y_size,0);
  
    for(size_type x=0;x<x_size;x++){
      for(size_type y=0;y<y_size;y++){

	bool found = false;
      	
	for(size_type z=min_bound(x,y)+1;(z<z_size);z++){
	  if ((visited_CENTER(x,y,z))&&(!found)){
	    found = true;
	    (*cut)(x,y) = z-1;
	  }
	}
      }
    }
  
  }


} // end of namepsace

#endif


